from __future__ import annotations

from typing import Any, Callable, Protocol, Sequence

from .utils import format_duration, format_time


class SchedulingPlanner:
    interval_strategy: IntervalStrategy

    def __init__(self, *, interval_strategy: IntervalStrategy = "start"):
        self.interval_strategy = interval_strategy

    def check(self, task: SchedulingTask, /) -> bool:
        return False

    def on_run(self, task: SchedulingTask, /) -> Any:
        return None

    def __str__(self) -> str:
        return "never"


class NeverPlanner(SchedulingPlanner):
    pass


class IntervalPlanner(SchedulingPlanner):
    def __init__(
        self,
        interval: float,
        *,
        interval_strategy: IntervalStrategy = "start",
    ):
        self.interval = interval
        self.interval_strategy = interval_strategy

    def check(self, task: SchedulingTask, /):
        if not self.interval:
            return not task.get_last_run(self.interval_strategy)

        return (task.get_last_run(self.interval_strategy) + self.interval) <= task.time

    def __str__(self):
        if not self.interval:
            return "once"

        return f"every {format_duration(self.interval)}"


class CooldownSource(Protocol):
    def get_cooldown(self, interval: float) -> float: ...


class CooldownPlanner(SchedulingPlanner):
    def __init__(
        self,
        cooldown_source: CooldownSource,
        interval: float,
        *,
        interval_strategy: IntervalStrategy = "start",
    ) -> None:
        self.cooldown_source = cooldown_source
        self.interval = interval
        self.interval_strategy = interval_strategy

    def check(self, task: SchedulingTask, /) -> bool:
        remaining = self.cooldown_source.get_cooldown(self.interval)
        return remaining <= 0

    def __str__(self):
        return f"every {format_duration(self.interval)} (cooling down using {self.cooldown_source})"


class FixedOffsetPlanner(SchedulingPlanner):
    """
    Planner that you can use to plan something at a fixed timestamp, optionally at intervals after.
    Keep in mind that `offset` is from unixtime 0, not from the program start although
    """

    def __init__(
        self,
        offset: float,
        interval: float,
        *,
        interval_strategy: IntervalStrategy = "start",
    ):
        self.offset = offset
        self.interval = interval
        self.interval_strategy = interval_strategy

    def get_period(self, ts: float) -> int:
        if not self.interval:
            return int(ts > self.offset)

        return int((ts - self.offset) // self.interval)

    def check(self, task: SchedulingTask, /) -> bool:
        last_run = task.get_last_run(self.interval_strategy)

        current_period = self.get_period(task.time)
        last_run_period = self.get_period(last_run)

        diff = current_period - last_run_period

        if diff > 1 and not last_run:
            # correct for the small starting offset
            # this will make the scheduler think that the last run was in this period,
            # even though there were no runs at all
            task.last_run_start = task.last_run_end = task.time
            diff = 0

        if diff > 0:
            print(last_run, current_period, last_run_period, diff)

        return diff > 0

    def __str__(self):
        start_fmt = format_time(self.offset)

        if not self.interval:
            return f"at {start_fmt}"

        return f"every {format_duration(self.interval)} since {start_fmt}"


class SwitchPlanner(SchedulingPlanner):
    """
    An advanced scheduler that switches between schedulers using a callback.
    """

    def __init__(
        self,
        *,
        planners: Sequence[SchedulingPlanner],
        switch_callback: Callable[[SchedulingTask], int],
        interval_strategy: IntervalStrategy = "start",
    ):
        self.planners = list(planners)
        self.planner_index = 0
        self.switch_callback = switch_callback

    @property
    def current_planner(self) -> SchedulingPlanner:
        return self.planners[self.planner_index % len(self.planners)]

    def check(self, task: SchedulingTask, /) -> bool:
        return self.current_planner.check(task)

    def on_run(self, task: SchedulingTask):
        self.current_planner.on_run(task)
        self.planner_index = self.switch_callback(task)


from .scheduler import IntervalStrategy
from .task import SchedulingTask



from __future__ import annotations

import inspect
from time import time
from asyncio import Task, AbstractEventLoop, get_event_loop
from time import time
from typing_extensions import Coroutine, Callable
from heapq import heapify
from random import randint


TaskCallback = Callable[..., Coroutine[None, None, None]]


class SchedulingTask:
    def __init__(
        self,
        callback: TaskCallback,
        planner: SchedulingPlanner,
        on_error: TaskCallback | None = None,
        name: str | None = None,
    ):
        self.callback = callback
        self.callback_kwargs = self.inspect_callback(callback)
        self.on_error = on_error

        self.planner = planner
        self.runs: dict[float, Task[None]] = {}
        self.scheduled: list[float] = [self.time]
        self.name = name

        self.last_run_start: float = 0
        self.last_run_end: float = 0

    def get_last_run(self, strategy: IntervalStrategy):
        if strategy == "start":
            return self.last_run_start

        if self.last_run_start > self.last_run_end:
            # runs now
            return self.time

        return self.last_run_end

    @property
    def time(self):
        return time()

    @staticmethod
    def inspect_callback(callback: TaskCallback) -> set[str]:
        signature = inspect.signature(callback)
        return set(signature.parameters)

    def run_callback(
        self,
        callback: TaskCallback,
        scheduler: Scheduler,
        event_loop: AbstractEventLoop | None,
    ):
        if not event_loop:
            event_loop = get_event_loop()

        kwargs = {
            "task": self,
            "scheduler": scheduler,
        }

        task = event_loop.create_task(
            callback(
                **{k: v for k, v in kwargs.items() if k in self.callback_kwargs},
            )
        )

        # save a reference to callback run so GC will not delete the task
        random_id = randint(0, int(10e9))
        self.runs[random_id] = task
        task.add_done_callback(lambda _: self.runs.pop(random_id, None))

    def tick(self, scheduler: Scheduler, event_loop: AbstractEventLoop | None = None):
        if self.planner.check(self):
            self.last_run_start = self.time
            self.run(self.callback, scheduler, event_loop)
            self.planner.on_run(self)
            self.last_run = self.time

    def run(self, callback: TaskCallback, scheduler: Scheduler, event_loop):
        async def callback_wrapper(**kwargs):
            try:
                await callback(**kwargs)
            except Exception as e:
                if self.on_error:
                    self.run_callback(self.on_error, scheduler, event_loop)
                else:
                    raise e

        self.run_callback(callback_wrapper, scheduler, event_loop)

    def discard(self):
        self.planner = NeverPlanner(interval_strategy="start")

    def pause_until(self, ts: float):
        def reinstate_switch(task: SchedulingTask):
            new_planner.planners.pop(new_planner.planner_index)
            self.planner = new_planner.planners[0]
            return 0

        paused = FixedOffsetPlanner(offset=ts, interval=0)

        new_planner = SwitchPlanner(
            planners=[self.planner, paused],
            switch_callback=reinstate_switch,
        )

        self.planner = new_planner

    def pause_for(self, duration: float):
        return self.pause_until(self.time + duration)

    @property
    def default_name(self):
        return self.callback.__name__

    @property
    def final_name(self):
        return self.name or self.default_name

    def __repr__(self):
        return f"Task['run {self.final_name} {self.planner}']"


from .scheduler import IntervalStrategy, Scheduler
from .planners import FixedOffsetPlanner, NeverPlanner, SchedulingPlanner, SwitchPlanner



units: dict[str, int] = {}

units["millisecond"] = 1
units["second"] = 1000
units["minute"] = 60 * units["second"]
units["hour"] = 60 * units["minute"]
units["day"] = 24 * units["hour"]
units["week"] = 7 * units["day"]


for unit in list(units):
    units[unit + "s"] = units[unit]


units["ms"] = units["millisecond"]
units["s"] = units["second"]
units["m"] = units["minute"]
units["h"] = units["hour"]
units["d"] = units["day"]
units["w"] = units["week"]


base_numbers = {
    k: i
    for i, k in enumerate(
        [
            "zero",
            "one",
            "two",
            "three",
            "four",
            "five",
            "six",
            "seven",
            "eight",
            "nine",
            "ten",
            "eleven",
            "twelve",
            "thirteen",
            "fourteen",
            "fiveteen",
            "sixteen",
            "seventeen",
            "eightteen",
            "nineteen",
        ]
    )
}

dozens = {
    k: (i + 2) * 10
    for i, k in enumerate(
        ["twenty", "thirty", "fourty", "fifty", "sixty", "seventy", "eighty", "ninety"]
    )
}

multipliers = {
    "hundred": 100,
    "hundreds": 100,
    "thousand": 1000,
    "thousands": 1000,
    "million": 1_000_000,
    "millions": 1_000_000,
    "billion": 1_000_000_000,
    "billions": 1_000_000_000,
}


def parse_interval(interval_string: str) -> float:
    parts = [part for part in interval_string.split("_") if part != "and"]
    result = 0

    number_buf: list[int] = []
    last_multiplier = False

    for part in parts:
        if part in units:
            amount = sum(number_buf) if number_buf else 1
            result += amount * units[part]
            number_buf.clear()

        elif part in multipliers:
            amount = multipliers[part]

            if number_buf:
                number_buf[-1] *= amount
            else:
                number_buf.append(amount)

        elif part in dozens:
            number_buf.append(dozens[part])

        elif part in base_numbers:
            amount = base_numbers[part]

            if number_buf and not last_multiplier:
                number_buf[-1] += amount
            else:
                number_buf.append(amount)

        else:
            continue

        last_multiplier = part in multipliers

    if number_buf:
        result += sum(number_buf)

    return result / 1000


def sum_units(data: dict[str, object]):
    result: float = 0

    for unit, value in data.items():
        if not isinstance(value, (float, int)):
            if unit in units:
                raise TypeError(
                    f"Value '{value}' (in '{unit}={value}') is not numeric. Pass a float or int to the scheduler policy"
                )
            else:
                continue

        if unit in units:
            result += units[unit] * value

    return result / 1000



from datetime import datetime

duration_sizes = {86400000: "d", 3600000: "h", 60000: "m", 1000: "s", 1: "ms"}


def format_duration(duration: float) -> str:
    parts: list[str] = []

    duration *= 1000

    for size, suffix in duration_sizes.items():
        amount = int(duration // size)

        if amount:
            parts.append(f"{amount}{suffix} ")

        duration %= size

    return "".join(parts).strip()


def format_time(ts: float) -> str:
    return datetime.fromtimestamp(ts).strftime("[%d.%m.%Y, %H:%M:%S, %:::z]")
